
import logging
from pathlib import Path
from subprocess import Popen, PIPE, CalledProcessError
from shutil import which
from typing import List


logger = logging.getLogger("rsync_runner")


def invoke_command(executable: Path, args: List[str], cwd: Path = Path().cwd(), ignore_errors: bool = False):
    stderr = ""
    try:
        if not executable.resolve().exists():
            raise FileNotFoundError(f"Could not find executable {str(executable)}.")
        cmd = f"{str(executable.resolve())} {' '.join(args)}"
        logger.info(cmd)
        proc = Popen(cmd, shell=True, cwd=str(cwd), stderr=PIPE, stdout=PIPE)
        while proc.poll() is None:
            _stdout, _stderr = proc.communicate()

            stderr = _stderr.decode('utf-8')
            if stderr and not ignore_errors:
                logger.error(stderr)

            stdout = _stdout.decode('utf-8')
            if stdout is not None:
                logger.info(stdout)

    except ValueError as e:
        logger.error(f"Invalid arguments: {str(e)}")

    except CalledProcessError as e:
        if not ignore_errors:
            logger.error(f"Error occurred with return code of {proc.returncode}.\nmsg: {e}\nstderr: {stderr}")


def enquote(string: str):
    return f'"{string}"'


def invoke_rsync(config):
    args = [enquote(config.source_dir), enquote(config.dest_dir)]
    options = ["--recursive", "--archive", "--human-readable"]

    if config.dry_run:
        options.append("--list-only")

    if config.write_report and config.report_path:
        report_path = config.log_directory_path.resolve().joinpath('rsyncrun.log')
        options.append(f"--log-file={str(enquote(report_path))}")
        options.append("--itemize-changes")

    if config.include_from:
        if isinstance(config.include_from, list):
            for file in config.include_from:
                if not file.exists():
                    raise FileNotFoundError(f"Include file '{file}' does not exist")
                options.append(f"--include-from={enquote(file)}")
        else:
            if not config.include_from.exists():
                raise FileNotFoundError(f"Include file '{config.include_from}' does not exist")
            options.append(f"--include-from={enquote(config.include_from)}")

    if config.exclude_from:
        if isinstance(config.exclude_from, list):
            for file in config.exclude_from:
                if not file.exists():
                    raise FileNotFoundError(f"Exclude file '{file}' does not exist")
                options.append(f"--exclude-from={enquote(file)}")
        else:
            if not config.exclude_from.exists():
                raise FileNotFoundError(f"Exclude file '{config.exclude_from}' does not exist")
            options.append(f"--exclude-from={enquote(config.exclude_from)}")

    if config.create_dirs:
        Path(config.dest_dir).mkdir(exist_ok=True, parents=True)

    options.extend(args)
    rsync = which("rsync")
    if not rsync:
        raise FileNotFoundError("rsync could not be found in PATH.")
    try:
        invoke_command(Path(rsync), options)
    except Exception:
        raise
